import os
import pandas as pd
import geopandas as gpd

def read_csv_spatial(filepath, type='dataframe', x_col=None, y_col=None, wkt_col=None, srid=None, **kwargs):
    """
    Read a CSV file and return as a pandas DataFrame, GeoPandas GeoDataFrame, or ESRI SEDF.

    Parameters
    ----------
    filepath : str
        Path to the CSV file.
    type : str, optional
        Output type: 'dataframe' (default), 'geodataframe', or 'sedf'.
    x_col : str or None, optional
        Name of the column to use for X (longitude) coordinates (for points).
    y_col : str or None, optional
        Name of the column to use for Y (latitude) coordinates (for points).
    wkt_col : str or None, optional
        Name of the column containing WKT geometries (for polygons).
    srid : int or str or None, optional
        EPSG code or CRS string to set for the geometry column.
    **kwargs :
        Additional keyword arguments passed to pandas.read_csv.

    Returns
    -------
    pandas.DataFrame, geopandas.GeoDataFrame, or arcgis.features.SpatialDataFrame
        The loaded data in the requested format.

    Raises
    ------
    ImportError
        If 'sedf' is requested but arcgis is not installed.
    ValueError
        If type is not recognized or required columns are missing.
    """
    df = pd.read_csv(filepath, **kwargs)
    if type == 'dataframe':
        return df
    elif type == 'geodataframe':
        if x_col and y_col and {x_col, y_col}.issubset(df.columns):
            gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df[x_col], df[y_col]))
        elif wkt_col and wkt_col in df.columns:
            gdf = gpd.GeoDataFrame(df, geometry=gpd.GeoSeries.from_wkt(df[wkt_col]))
        else:
            raise ValueError("Cannot create GeoDataFrame: provide x_col and y_col for points, or wkt_col for polygons.")
        if srid:
            gdf = gdf.set_crs(srid)
        return gdf
    elif type == 'sedf':
        try:
            import arcgis
        except ImportError:
            raise ImportError("arcgis is required for SEDF output. Please install arcgis package.")
        if x_col and y_col and {x_col, y_col}.issubset(df.columns):
            sdf = df.spatial.from_xy(df,x_column=x_col, y_column=y_col, sr=srid)
        elif wkt_col and wkt_col in df.columns:
            sdf = df.spatial.from_wkt(df,wkt_col, sr=srid)
        else:
            raise ValueError("Cannot create SEDF: provide x_col and y_col for points, or wkt_col for polygons.")
        return sdf
    else:
        raise ValueError(f"Unknown type: {type}. Must be one of 'dataframe', 'geodataframe', or 'sedf'.")

def get_detroit_911_calls(type='dataframe'):
    """
    Load the Detroit 911 Calls dataset from the data directory.

    Parameters
    ----------
    type : str, optional
        Output type: 'dataframe' (default), 'geodataframe', or 'sedf'.


    Returns
    -------
    pandas.DataFrame, geopandas.GeoDataFrame, or arcgis.features.SpatialDataFrame
        The loaded data in the requested format.
    """
    
    data_path = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'data', 'Detroit 911 Calls.csv')
    # Always use longitude and latitude for spatial output
    if type in ('geodataframe', 'sedf'):
        return read_csv_spatial(data_path, type=type, x_col='longitude', y_col='latitude')
    else:
        return read_csv_spatial(data_path, type=type)
