import numpy as np
from scipy.stats import norm
from scipy.spatial.distance import cdist
from .spatial_relationships import knn_weight_matrix
from .spatial_relationships import distance_band_weight_matrix
from .spatial_relationships import grid_contiguity_weight_matrix


def calculate_pvalue(observed_statistic, shuffled_statistics):
    """
    Calculate the p-value of an observed statistic using shuffled statistics.
    
    Parameters:
        observed_statistic (float): The observed value of the statistic.
        shuffled_statistics (np.array): Array of statistics calculated from shuffled data.
    
    Returns:
        pvalue (float): The p-value.
    """
    # Calculate the p-value using SciPy's percentileofscore
    z_score = (observed_statistic - np.array(shuffled_statistics).mean() ) / np.array(shuffled_statistics).std()
    p_value = 2 * (1 - norm.cdf(np.abs(z_score)))
    return p_value


def test_spatial_index_significance(sdf, variable, index_func,permutations=999, seed=None,**kwargs):

    """
    Generic significance test for any spatial index using Monte Carlo permutations.

    Parameters
    ----------
    sdf : pandas.DataFrame
        Spatially enabled DataFrame with 'SHAPE' column.
    variable : str
        Name of the variable to analyze.
    index_func : callable
        Function to compute the spatial index. Must accept sdf, variable, and **kwargs.
    permutations : int
        Number of permutations to perform.
    seed : int or None
        Seed for reproducibility.
    **kwargs : dict
        Keyword arguments to pass to the index function.

    Returns
    -------
    dict
        {
            'observed': float,
            'permuted': np.ndarray,
            'p_value': float
        }
    """
    if seed is not None:
        np.random.seed(seed)

    # Compute the observed index value
    observed = index_func(sdf, variable, **kwargs)

    original_values = sdf[variable].values.copy()

    # List comprehension replaces loop
    permuted_stats = np.array([
        index_func(sdf.assign(_permuted_=np.random.permutation(original_values)), '_permuted_', **kwargs)
        for _ in range(permutations)
    ])

    # Compute p-value
    p_value =  calculate_pvalue(observed, permuted_stats) 

    return {
        'observed': observed,
        'permuted': permuted_stats,
        'p_value': p_value
    }


def spatial_variance_ratio(sdf, variable, weights=None, method='knn', 
                          k=4, threshold=None, cell_size=None, contiguity="rook",
                          row_standardized=False, symmetric=True):
    """
    Calculate the Spatial Variance Ratio (SVR) using various spatial weight matrix methods.

    Parameters
    ----------
    sdf : pandas.DataFrame
        Spatially enabled DataFrame with 'SHAPE' column containing geometry objects.
    variable : str
        Name of the variable to analyze.
    weights : scipy.sparse.csr_matrix or None, optional
        Pre-computed spatial weights matrix. If None, weights will be calculated.
    method : str, optional
        Method for constructing spatial weights:
        - 'knn': K-nearest neighbors
        - 'distance_band': Fixed distance band
        - 'grid': Grid-based contiguity
        Default is 'knn'.
    k : int, optional
        Number of nearest neighbors for KNN method (default is 4).
    threshold : float, optional
        Distance threshold for distance band method.
    cell_size : float, optional
        Grid cell size for grid contiguity method.
    contiguity : str, optional
        Type of contiguity for grid method ('rook' or 'queen', default is 'rook').
    row_standardized : bool, optional
        Whether to row-standardize the weight matrix (default is False).
    symmetric : bool, optional
        Whether to ensure symmetry in the weight matrix (default is True).

    Returns
    -------
    float
        Spatial Variance Ratio (SVR) statistic. Returns np.nan if calculation fails.

    Notes
    -----
    The Spatial Variance Ratio compares the average local variance (within neighborhoods)
    to the global variance. Values:
    - > 1: Higher local variance than global (negative spatial dependence)
    - ≈ 1: No spatial structure
    - < 1: Lower local variance than global (positive spatial dependence)
    """
    coords = np.array([[geom.x, geom.y] for geom in sdf['SHAPE']])
    values = np.asarray(sdf[variable].values)
    n = len(values)
    
    if weights is None:
        if method == 'knn':
            W = knn_weight_matrix(coords, k=k, row_standardized=row_standardized,
                                return_sparse=True, symmetric=symmetric)
        elif method == 'distance_band':
            if threshold is None:
                raise ValueError("threshold must be specified for distance_band method")
            W = distance_band_weight_matrix(coords, threshold=threshold,
                                          row_standardized=row_standardized,
                                          return_sparse=True)
        elif method == 'grid':
            if cell_size is None:
                raise ValueError("cell_size must be specified for grid method")
            W = grid_contiguity_weight_matrix(coords, cell_size=cell_size,
                                            contiguity=contiguity,
                                            row_standardized=row_standardized,
                                            return_sparse=True)
        else:
            raise ValueError(f"Method '{method}' not supported. Use 'knn', 'distance_band', or 'grid'")
    else:
        W = weights
        
    # Compute local variances
    local_variances = []
    for i in range(n):
        neighbors = W[i].nonzero()[1]
        if len(neighbors) == 0:
            continue
        neighbor_values = values[neighbors]
        var = np.var(neighbor_values, ddof=1)
        local_variances.append(var)

    if not local_variances:
        return np.nan

    avg_local_variance = np.mean(local_variances)
    global_variance = np.var(values, ddof=1)

    svr = avg_local_variance / global_variance if global_variance != 0 else np.nan
    return svr


def morans_i(sdf, variable, weights=None, method='distance_band',
             k=4, threshold=None, cell_size=None, contiguity="rook",
             row_standardized=True, symmetric=True):
    """
    Calculate Moran's I statistic for spatial autocorrelation using various weight matrix methods.

    Parameters
    ----------
    sdf : pandas.DataFrame
        Spatially enabled DataFrame with 'SHAPE' column containing geometry objects.
    variable : str
        Name of the variable to analyze.
    weights : scipy.sparse.csr_matrix or None, optional
        Pre-computed spatial weights matrix. If None, weights will be calculated.
    method : str, optional
        Method for constructing spatial weights:
        - 'knn': K-nearest neighbors
        - 'distance_band': Fixed distance band
        - 'grid': Grid-based contiguity
        Default is 'distance_band'.
    k : int, optional
        Number of nearest neighbors for KNN method (default is 4).
    threshold : float, optional
        Distance threshold for distance band method.
    cell_size : float, optional
        Grid cell size for grid contiguity method.
    contiguity : str, optional
        Type of contiguity for grid method ('rook' or 'queen', default is 'rook').
    row_standardized : bool, optional
        Whether to row-standardize the weight matrix (default is True for Moran's I).
    symmetric : bool, optional
        Whether to ensure symmetry in the weight matrix (default is True).

    Returns
    -------
    float
        Moran's I statistic.

    Raises
    ------
    ValueError
        If all values are identical (zero variance) or if required parameters are missing.

    Notes
    -----
    Moran's I interpretation:
    - I > E[I]: Positive spatial autocorrelation
    - I ≈ E[I]: No spatial autocorrelation
    - I < E[I]: Negative spatial autocorrelation
    Expected value E[I] = -1/(n-1) where n is number of observations
    """
    coords = np.array([[geom.x, geom.y] for geom in sdf['SHAPE']])
    X = np.asarray(sdf[variable].values)
    n = X.shape[0]
    
    if weights is None:
        if method == 'knn':
            W = knn_weight_matrix(coords, k=k, row_standardized=row_standardized,
                               return_sparse=True, symmetric=symmetric)
        elif method == 'distance_band':
            if threshold is None:
                raise ValueError("threshold must be specified for distance_band method")
            W = distance_band_weight_matrix(coords, threshold=threshold,
                                          row_standardized=row_standardized,
                                          return_sparse=True)
        elif method == 'grid':
            if cell_size is None:
                raise ValueError("cell_size must be specified for grid method")
            W = grid_contiguity_weight_matrix(coords, cell_size=cell_size,
                                            contiguity=contiguity,
                                            row_standardized=row_standardized,
                                            return_sparse=True)
        else:
            raise ValueError(f"Method '{method}' not supported. Use 'knn', 'distance_band', or 'grid'")
    else:
        W = weights
    
    # Check for zero variance case
    if np.std(X) == 0:
        raise ValueError("All values are identical. Moran's I cannot be calculated.")
    
    # Calculate Moran's I
    X_mean = np.mean(X)
    X_diff = X - X_mean
    numerator = np.sum(W.multiply(np.outer(X_diff, X_diff)))
    denominator = np.sum(X_diff ** 2)
    S0 = np.sum(W)
    
    I = (n / S0) * (numerator / denominator)
    return I
